# -*- coding: utf-8 -*-
"""Vectorized two diode model."""

import numpy as np
from scipy import constants
from .pvmismatch import pvconstants, pvcell

# ------------------------------------------------------------------------------
# TWO DIODE MODEL -------------------------------------------------------------


# PVConstants
k = constants.k  #: [kJ/mole/K] Boltzmann constant
q = constants.e  #: [Coloumbs] elementary charge
E0 = 1000.  #: [W/m^2] irradiance of 1 sun
T0 = 298.15  #: [K] reference temperature
# Other Constants
EPS = np.finfo(np.float64).eps
VBYPASS = np.float64(-0.5)  # [V] trigger voltage of bypass diode
CELLAREA = np.float64(153.33)  # [cm^2] cell area


def two_diode_model(pvcs, Ee, u_cell_type, Tcell, NPTS=200, NPTS_cell=100, use_cell_NPT=False):
    """
    Estimate IV curve using the two diode model native to pvmismatch.

    Parameters
    ----------
    pvcs : list
        List of different pvmismatch pvcell objects. Most modules have the same cells. Some might not.
    Ee : numpy.ndarray
        1-D array containing irradiances in suns.
    u_cell_type : list
        List of cell types at each irradiance setting.
    Tcell : numpy.ndarray
        1-D array containing cell temperatures in Kelvin.
    NPTS : int, optional
        Number of points in IV curve. The default is 200.
    NPTS_cell : int, optional
        Number of points in cell IV curve. The default is 100.
    use_cell_NPT : bool, optional
        Turn on this if cell IV curves use 'NPTS_cell'. The default is False.

    Returns
    -------
    cell_data : dict
        Dictionary containing cell IV curves.

    """
    # Generate NPTs
    pts, negpts, Imod_pts, Imod_negpts = NPTS_f(
        NPTS, np.repeat(Ee[:, np.newaxis], NPTS, axis=1).shape)
    if not use_cell_NPT:
        NPTS_cell = NPTS

    Icell_vec = np.zeros((len(Ee), NPTS_cell*3))
    Vcell_vec = Icell_vec.copy()
    Pcell_vec = Icell_vec.copy()
    VRBD_vec = np.zeros((len(Ee), 1))
    Voc_vec = VRBD_vec.copy()
    Isc_vec = VRBD_vec.copy()
    for ie, ee in enumerate(Ee):
        uct = int(u_cell_type[ie])
        pvc = pvcs[uct]
        # Define cell parameters
        Rs = pvc.Rs
        Rsh = pvc.Rsh
        Isat1_T0 = pvc.Isat1_T0
        Isat2_T0 = pvc.Isat2_T0
        Isc0_T0 = pvc.Isc0_T0
        aRBD = pvc.aRBD
        bRBD = pvc.bRBD
        VRBD = pvc.VRBD
        nRBD = pvc.nRBD
        Eg = pvc.Eg
        alpha_Isc = pvc.alpha_Isc
        # PVMismatch
        pvconst = pvconstants.PVconstants(npts=NPTS_cell)
        # create pvcell
        pvc = pvcell.PVcell(pvconst=pvconst, Rs=Rs,
                            Rsh=Rsh,
                            Isat1_T0=Isat1_T0,
                            Isat2_T0=Isat2_T0,
                            Isc0_T0=Isc0_T0,
                            aRBD=aRBD,
                            bRBD=bRBD,
                            VRBD=VRBD,
                            nRBD=nRBD,
                            Eg=Eg,
                            alpha_Isc=alpha_Isc)
        pvc.Ee = ee
        pvc.Tcell = Tcell[ie]
        Icell_vec[ie, :] = pvc.Icell.flatten()
        Vcell_vec[ie, :] = pvc.Vcell.flatten()
        Pcell_vec[ie, :] = pvc.Pcell.flatten()
        VRBD_vec[ie, 0] = pvc.VRBD
        Voc_vec[ie, 0] = pvc.Voc
        Isc_vec[ie, 0] = pvc.Isc

    # Store in a dict
    cell_data = dict()
    cell_data['Icell'] = Icell_vec
    cell_data['Vcell'] = Vcell_vec
    cell_data['Pcell'] = Pcell_vec
    cell_data['VRBD'] = VRBD_vec
    cell_data['Voc'] = Voc_vec
    cell_data['Isc'] = Isc_vec
    cell_data['NPT'] = dict()
    cell_data['NPT']['pts'] = pts
    cell_data['NPT']['negpts'] = negpts
    cell_data['NPT']['Imod_pts'] = Imod_pts
    cell_data['NPT']['Imod_negpts'] = Imod_negpts
    cell_data['NPT']['Npts'] = NPTS

    return cell_data


def NPTS_f(Npts=200, vec_shape=(1, 1, 1, 1)):
    """
    Generate point spacing for IV curves for both cells and module.

    Parameters
    ----------
    Npts : int, optional
        Number of points in IV curve. The default is 200. The default is 200.
    vec_shape : tuple, optional
        tuple with system shape (num_string, string_len, row_cells, col_cells). The default is (1, 1, 1, 1).

    Returns
    -------
    pts_vec : numpy.ndarray
        Positive point spacing for cell IV curves.
    negpts_vec : numpy.ndarray
        Negative point spacing for cell IV curves.
    Imod_pts_vec : numpy.ndarray
        Positive point spacing for module IV curves.
    Imod_negpts_vec : numpy.ndarray
        Negative point spacing for module IV curves.

    """
    # set number of points in IV curve(s)
    # point spacing from 0 to 1, used for Vcell, Vmod, Vsys and Istring
    # decrease point spacing as voltage approaches Voc by using logspace
    pts_vec = np.zeros(vec_shape)
    negpts_vec = pts_vec.copy()
    Imod_pts_vec = pts_vec.copy()
    Imod_negpts_vec = pts_vec.copy()
    pts = (11. - np.logspace(np.log10(11.), 0., Npts)) / 10.
    pts[0] = 0.  # first point must be exactly zero
    pts_vec[:, :] = pts
    negpts = (11. - np.logspace(np.log10(11. - 1. / float(Npts)),
                                0., Npts)) / 10.
    Imod_negpts = 1 + 1. / float(Npts) / 10. - negpts
    Imod_negpts_vec[:, :] = Imod_negpts
    negpts = np.flipud(negpts)  # reverse the order
    negpts_vec[:, :] = negpts
    # shift and concatenate pvconst.negpts and pvconst.pts
    # so that tight spacing is around MPP and RBD
    Imod_pts = 1 - np.flipud(pts)
    Imod_pts_vec[:, :] = Imod_pts

    return pts_vec, negpts_vec, Imod_pts_vec, Imod_negpts_vec



# -*- coding: utf-8 -*-
"""
This module defines the :class:`~pvmismatch.pvmismatch_lib.pvmodule.PVmodule`.
"""

from __future__ import absolute_import
from past.builtins import xrange, range
from builtins import zip
from six import itervalues
import numpy as np
from copy import copy
from matplotlib import pyplot as plt
# use absolute imports instead of relative, so modules are portable
from .pvconstants import PVconstants, get_series_cells
from .pvcell import PVcell
from .pvexceptions import PVexception

VBYPASS = np.float64(-0.5)  # [V] trigger voltage of bypass diode
CELLAREA = np.float64(153.33)  # [cm^2] cell area
DEFAULT_BYPASS = 0
MODULE_BYPASS = 1
CUSTOM_SUBSTR_BYPASS = 2


def standard_cellpos_pat(nrows, ncols_per_substr):
    """
    Standard module object builder.

    Parameters
    ----------
    nrows : integer
        Number of rows of cells in module.

    ncols_per_substr : list of integers
        Number of columns of cells in each substring (in parallel with a diode).

    Returns
    -------
    cellpos : list of lists
         Outermost list is a list of substrings in parallel with a bypass diode
         and in series with each other.
         The substring is a list of columns in that substring.
         Inside the column are the actual cells in each row.
         Each cell has a 'crosstie' key and an index.
    """
    cellpos = []
    ncols = [0, 0]
    for substr_cols in ncols_per_substr:
        ncols[0], ncols[1] = ncols[1], ncols[1] + substr_cols
        newsubstr = []
        for col in xrange(*ncols):
            newrow = []
            for row in xrange(nrows):
                idx = col * nrows
                if col % 2 == 0:
                    idx += row
                else:
                    idx += (nrows - row - 1)
                newrow.append({'crosstie': False, 'idx': idx})
            newsubstr.append(newrow)
        cellpos.append(newsubstr)
    return cellpos


# standard cell positions presets
STD24 = standard_cellpos_pat(1, [1] * 24)
STD72 = standard_cellpos_pat(12, [2, 2, 2])
STD96 = standard_cellpos_pat(12, [2, 4, 2])
"""
Standard module: 12x8 cells
Substrings have 2, 4 and 2 columns of cells per diode
"""
STD128 = standard_cellpos_pat(16, [2, 4, 2])


def crosstied_cellpos_pat(nrows_per_substrs, ncols, partial=False):
    """
    Cross-tied module object builder.

    Parameters
    ----------
    nrows_per_substr : list of integers
        Number of rows of cells in each substring (in parallel with a diode).

    ncols : integer
        Number of columns of cells.

    partial : boolean
        False (default) means TCT (all cells are cross-tied).
        True means no cross-tiling.

    Returns
    -------
    cellpos : list of lists
         Outermost list is a list of substrings in parallel with a bypass diode
         and in series with each other.
         The substring is a list of columns in that substring.
         Inside the column are the actual cells in each row.
         Each cell has a 'crosstie' key and an index.
    """
    trows = sum(nrows_per_substrs)
    cellpos = []
    nrows = [0, 0]
    for substr_row in nrows_per_substrs:
        nrows[0], nrows[1] = nrows[1], nrows[1] + substr_row
        newsubstr = []
        for col in xrange(ncols):
            newrow = []
            for row in xrange(*nrows):
                crosstie = True
                if partial and newrow:
                    crosstie = False
                newrow.append({'crosstie': crosstie, 'idx': col * trows + row})
            newsubstr.append(newrow)
        cellpos.append(newsubstr)
    return cellpos


# crosstied cell positions presets
TCT492 = crosstied_cellpos_pat([27, 28, 27], 6)
"""
Standard Tiled module with total cross-ties: 82x6 cells
Substrings have 27, 28 and 27 rows of cells per diode
"""
PCT492 = crosstied_cellpos_pat([27, 28, 27], 6, partial=True)
"""
Standard Tiled module with partial cross-ties: 82x6 cells
Substrings have 27, 28 and 27 rows of cells per diode
"""


def combine_parallel_circuits(IVprev_cols, pvconst):
    """
    Combine crosstied circuits in a substring

    :param IVprev_cols: lists of IV curves of crosstied and series circuits
    :param pvconst: an instance of :class:`~pvmismatch.pvconstants.PVconstants`
    :return: current [A] and voltage [V] of the combined parallel circuites
    """
    # combine crosstied circuits
    Irows, Vrows = [], []
    Isc_rows, Imax_rows = [], []
    for IVcols in zip(*IVprev_cols):
        Iparallel, Vparallel, Voc_parallel = zip(*IVcols)
        Iparallel = np.asarray(Iparallel)
        Vparallel = np.asarray(Vparallel)
        Voc_parallel = np.mean(Voc_parallel)
        Irow, Vrow = pvconst.calcParallel(
            Iparallel, Vparallel, Vparallel.max(),
            Vparallel.min(), Voc=Voc_parallel
        )
        Irows.append(Irow)
        Vrows.append(Vrow)
        Isc_rows.append(np.interp(np.float64(0), Vrow, Irow))
        Imax_rows.append(Irow.max())
    Irows, Vrows = np.asarray(Irows), np.asarray(Vrows)
    Isc_rows = np.asarray(Isc_rows)
    Imax_rows = np.asarray(Imax_rows)
    return pvconst.calcSeries(
        Irows, Vrows, Isc_rows.mean(), Imax_rows.max()
    )


def parse_diode_config(Vbypass, cell_pos):
    """
    Parse diode configuration from the Vbypass argument
    :param Vbypass: Vbypass config
    :type Vbypass: float|list|tuple
    :param cell_pos:
    :type cell_pos:
    :return: bypass config
    :rtype: str
    """
    try:
        # check if float or list/tuple
        num_bypass = len(Vbypass)
    except TypeError:
        # float passed - default case - Vbypass across every cell string
        return DEFAULT_BYPASS
    else:
        # if only one value is passed in the list- assume only one
        # bypass diode  across the PV module
        if len(Vbypass) == 1:
            return MODULE_BYPASS
        # if more than 1 values are passed, apply them across
        # the cell strings in ascending order
        elif len(cell_pos) == num_bypass:
            return CUSTOM_SUBSTR_BYPASS
        else:
            raise PVexception("wrong number of bypass diode values passed : %d" % (len(Vbypass)))


class PVmodule(object):
    """
    A Class for PV modules.

    :param cell_pos: cell position pattern
    :type cell_pos: dict
    :param pvcells: list of :class:`~pvmismatch.pvmismatch_lib.pvcell.PVcell`
    :type pvcells: list, :class:`~pvmismatch.pvmismatch_lib.pvcell.PVcell`
    :param pvconst: An object with common parameters and constants.
    :type pvconst: :class:`~pvmismatch.pvmismatch_lib.pvconstants.PVconstants`
    :param Vbypass: float|list of :float
        bypass diode trigger voltage [V]
        default case - one bypass diode per cell string (VBYPASS = -0.5V(V)) \n
        float - one bypass diode per cell string with Vf = Vbypass (V) \n
        len(list) == 1 - one bypass diode per module (bypasses entire module ) \n
        len(list) == len(cell_pos) - bypass diode value across cell string as defined in the list \n
    :param cellArea: cell area [cm^2]
    """

    def __init__(self, cell_pos=STD96, pvcells=None, pvconst=None,
                 Vbypass=None, cellArea=CELLAREA):
        # TODO: check cell position pattern
        self.cell_pos = cell_pos  #: cell position pattern dictionary
        self.numberCells = sum([len(c) for s in self.cell_pos for c in s])
        """number of cells in the module"""
        # is pvcells a list?
        try:
            pvc0 = pvcells[0]
        except TypeError:
            # is pvcells an object?
            try:
                pvconst = pvcells.pvconst
            except AttributeError:
                #  try to use the pvconst arg or create one if none
                if not pvconst:
                    pvconst = PVconstants()
                # create pvcell
                pvcells = PVcell(pvconst=pvconst)
            # expand pvcells to list
            pvcells = [pvcells] * self.numberCells
        else:
            pvconst = pvc0.pvconst
            for p in pvcells:
                if p.pvconst is not pvconst:
                    raise Exception('PVconstant must be the same for all cells')
        self.pvconst = pvconst  #: configuration constants

        # set default value of Vbypass if None
        if Vbypass is None:
            self.Vbypass = VBYPASS  #: [V] trigger voltage of bypass diode
        else:
            # if an object is passed, use that to determine the config of bypass diodes
            self.Vbypass = Vbypass
        self.Vbypass_config = parse_diode_config(self.Vbypass, self.cell_pos)

        self.cellArea = cellArea  #: [cm^2] cell area
        # check cell position pattern matches list of cells
        if len(pvcells) != self.numberCells:
            raise ValueError(
                "Number of cells doesn't match cell position pattern."
            )
        self.pvcells = pvcells  #: list of `PVcell` objects in this `PVmodule`
        self.numSubStr = len(self.cell_pos)  #: number of substrings
        self.subStrCells = [len(_) for _ in self.cell_pos]  #: cells per substr
        # initialize members so PyLint doesn't get upset
        self.Imod, self.Vmod, self.Pmod, self.Isubstr, self.Vsubstr = self.calcMod()

    # TODO: use __getattr__ to check for updates to pvcells

    # copy some values from cells to modules
    @property
    def Ee(self):
        return np.array([pvc.Ee.flatten() for pvc in self.pvcells])

    @property
    def Tcell(self):
        return np.array([pvc.Tcell.flatten() for pvc in self.pvcells])

    @property
    def Icell(self):
        return np.array([pvc.Icell.flatten() for pvc in self.pvcells])

    @property
    def Vcell(self):
        return np.array([pvc.Vcell.flatten() for pvc in self.pvcells])

    @property
    def Pcell(self):
        return np.array([pvc.Pcell.flatten() for pvc in self.pvcells])

    @property
    def Isc(self):
        return np.array([pvc.Isc.flatten() for pvc in self.pvcells])

    @property
    def Voc(self):
        return np.array([pvc.Voc.flatten() for pvc in self.pvcells])

    @property
    def VRBD(self):
        return np.array([pvc.VRBD.flatten() for pvc in self.pvcells])

    def setSuns(self, Ee, cells=None):
        """
        Set the irradiance in suns, Ee, on the solar cells in the module.
        Recalculates cell current (Icell [A]), voltage (Vcell [V]) and power
        (Pcell [W]) as well as module current (Imod [A]), voltage (Vmod [V])
        and power (Pmod [W]).

        Args:
            Ee (:class:`numpy.ndarray`): Effective Irradiance [suns]
            cells (list): Cells to change [Optional]
        """
        if cells is None:
            if np.isscalar(Ee):
                new_pvcells = range(self.numberCells)  # new list of cells
                old_pvcells = dict.fromkeys(self.pvcells)  # same as set(pvcells)
                for cell_id, pvcell in enumerate(self.pvcells):
                    if old_pvcells[pvcell] is None:
                        new_pvcells[cell_id] = copy(pvcell)
                        old_pvcells[pvcell] = new_pvcells[cell_id]
                    else:
                        new_pvcells[cell_id] = old_pvcells[pvcell]
                self.pvcells = new_pvcells
                pvcell_set = itervalues(old_pvcells)
                for pvc in pvcell_set:
                    pvc.Ee = Ee
            elif np.size(Ee) == self.numberCells:
                self.pvcells = copy(self.pvcells)  # copy list first
                for cell_idx, Ee_idx in enumerate(Ee):
                    self.pvcells[cell_idx] = copy(self.pvcells[cell_idx])
                    self.pvcells[cell_idx].Ee = Ee_idx
            else:
                raise Exception("Input irradiance value (Ee) for each cell!")
        else:
            Ncells = np.size(cells)
            self.pvcells = copy(self.pvcells)  # copy list first
            if np.isscalar(Ee):
                cells_to_update = [self.pvcells[i] for i in cells]
                old_pvcells = dict.fromkeys(cells_to_update)
                for cell_id, pvcell in zip(cells, cells_to_update):
                    if old_pvcells[pvcell] is None:
                        self.pvcells[cell_id] = copy(pvcell)
                        self.pvcells[cell_id].Ee = Ee
                        old_pvcells[pvcell] = self.pvcells[cell_id]
                    else:
                        self.pvcells[cell_id] = old_pvcells[pvcell]
            elif np.size(Ee) == Ncells:
                # Find unique irradiance values
                # TODO possible "cleaner" alternative by grouping cells into tuples that match the set irradiance
                # E.g: pvsys.setSuns({X: {Y: {'Ee': (0.33, 0.99), 'cells': [(2, 3), 17]}}})
                cells = np.array(cells)
                Ee = np.array(Ee)
                unique_ee = np.unique(Ee)
                for a_Ee in unique_ee:
                    cells_subset = cells[Ee == a_Ee]
                    cells_to_update = [self.pvcells[i] for i in cells_subset]
                    old_pvcells = dict.fromkeys(cells_to_update)
                    for cell_id, pvcell in zip(cells_subset, cells_to_update):
                        if old_pvcells[pvcell] is None:
                            self.pvcells[cell_id] = copy(pvcell)
                            self.pvcells[cell_id].Ee = a_Ee
                            old_pvcells[pvcell] = self.pvcells[cell_id]
                        else:
                            self.pvcells[cell_id] = old_pvcells[pvcell]
            else:
                raise Exception("Input irradiance value (Ee) for each cell!")
        self.Imod, self.Vmod, self.Pmod, self.Isubstr, self.Vsubstr = self.calcMod()

    # TODO setTemps is a nearly identical copy of setSuns. The DRY principle
    # says that we should not be copying code.
    # TODO Replace both setSuns() and setTemps() with a single method for
    # updating cell parameters that works for all params

    def setTemps(self, Tc, cells=None):
        """
        Set the temperature in Kelvin, Tc, on the solar cells in the module.
        Recalculates cell current (Icell [A]), voltage (Vcell [V]) and power
        (Pcell [W]) as well as module current (Imod [A]), voltage (Vmod [V])
        and power (Pmod [W]).

        Args:
            Tc (:class:`numpy.ndarray`): Cell Temperature [K]
            cells (list): Cells to change [Optional]
        """
        if cells is None:
            if np.isscalar(Tc):
                new_pvcells = range(self.numberCells)  # new list of cells
                old_pvcells = dict.fromkeys(self.pvcells)  # same as set(pvcells)
                for cell_id, pvcell in enumerate(self.pvcells):
                    if old_pvcells[pvcell] is None:
                        new_pvcells[cell_id] = copy(pvcell)
                        old_pvcells[pvcell] = new_pvcells[cell_id]
                    else:
                        new_pvcells[cell_id] = old_pvcells[pvcell]
                self.pvcells = new_pvcells
                pvcell_set = itervalues(old_pvcells)
                for pvc in pvcell_set:
                    pvc.Tcell = Tc
            elif np.size(Tc) == self.numberCells:
                self.pvcells = copy(self.pvcells)  # copy list first
                for cell_idx, Tc_idx in enumerate(Tc):
                    self.pvcells[cell_idx] = copy(self.pvcells[cell_idx])
                    self.pvcells[cell_idx].Tcell = Tc_idx
            else:
                raise Exception("Input temperature value (Tc) for each cell!")
        else:
            Ncells = np.size(cells)
            self.pvcells = copy(self.pvcells)  # copy list first
            if np.isscalar(Tc):
                cells_to_update = [self.pvcells[i] for i in cells]
                old_pvcells = dict.fromkeys(cells_to_update)
                for cell_id, pvcell in zip(cells, cells_to_update):
                    if old_pvcells[pvcell] is None:
                        self.pvcells[cell_id] = copy(pvcell)
                        self.pvcells[cell_id].Tcell = Tc
                        old_pvcells[pvcell] = self.pvcells[cell_id]
                    else:
                        self.pvcells[cell_id] = old_pvcells[pvcell]
            elif np.size(Tc) == Ncells:
                # Find unique irradiance values
                # TODO possible "cleaner" alternative by grouping cells into tuples that match the set temp
                # E.g: pvsys.setTemps({X: {Y: {'Tc': (280, 290), 'cells': [(2, 3), 17]}}})
                cells = np.array(cells)
                Tc = np.array(Tc)
                unique_tc = np.unique(Tc)
                for a_Tc in unique_tc:
                    cells_subset = cells[Tc == a_Tc]
                    cells_to_update = [self.pvcells[i] for i in cells_subset]
                    old_pvcells = dict.fromkeys(cells_to_update)
                    for cell_id, pvcell in zip(cells_subset, cells_to_update):
                        if old_pvcells[pvcell] is None:
                            self.pvcells[cell_id] = copy(pvcell)
                            self.pvcells[cell_id].Tcell = a_Tc
                            old_pvcells[pvcell] = self.pvcells[cell_id]
                        else:
                            self.pvcells[cell_id] = old_pvcells[pvcell]
            else:
                raise Exception("Input temperature value (Tc) for each cell!")
        self.Imod, self.Vmod, self.Pmod, self.Isubstr, self.Vsubstr = self.calcMod()

    def calcMod(self):
        """
        Calculate module I-V curves.

        Returns module currents [A], voltages [V] and powers [W]
        """
        # iterate over substrings
        # TODO: benchmark speed difference append() vs preallocate space
        Isubstr, Vsubstr, Isc_substr, Imax_substr = [], [], [], []
        for substr_idx, substr in enumerate(self.cell_pos):
            # check if cells are in series or any crosstied circuits
            if all(r['crosstie'] == False for c in substr for r in c):
                idxs = [r['idx'] for c in substr for r in c]
                IatVrbd = np.asarray(
                    [np.interp(vrbd, v, i) for vrbd, v, i in
                     zip(self.VRBD[idxs], self.Vcell[idxs], self.Icell[idxs])]
                )
                Isub, Vsub = self.pvconst.calcSeries(
                    self.Icell[idxs], self.Vcell[idxs], self.Isc[idxs].mean(),
                    IatVrbd.max()
                )
            elif all(r['crosstie'] == True for c in substr for r in c):
                Irows, Vrows = [], []
                Isc_rows, Imax_rows = [], []
                for row in zip(*substr):
                    idxs = [c['idx'] for c in row]
                    Irow, Vrow = self.pvconst.calcParallel(
                        self.Icell[idxs], self.Vcell[idxs],
                        self.Voc[idxs].max(), self.VRBD.min(),
                        Voc=self.Voc[idxs].mean()
                    )
                    Irows.append(Irow)
                    Vrows.append(Vrow)
                    Isc_rows.append(np.interp(np.float64(0), Vrow, Irow))
                    Imax_rows.append(Irow.max())
                Irows, Vrows = np.asarray(Irows), np.asarray(Vrows)
                Isc_rows = np.asarray(Isc_rows)
                Imax_rows = np.asarray(Imax_rows)
                Isub, Vsub = self.pvconst.calcSeries(
                    Irows, Vrows, Isc_rows.mean(), Imax_rows.max()
                )
            else:
                IVall_cols = []
                prev_col = None
                IVprev_cols = []
                for col in substr:
                    IVcols = []
                    is_first = True
                    # combine series between crossties
                    for idxs in get_series_cells(col, prev_col):
                        if not idxs:
                            # first row should always be empty since it must be
                            # crosstied
                            is_first = False
                            continue
                        elif is_first:
                            # TODO: use pvmismatch exceptions
                            raise Exception(
                                "First row and last rows must be crosstied."
                            )
                        elif len(idxs) > 1:
                            IatVrbd = np.asarray(
                                [np.interp(vrbd, v, i) for vrbd, v, i in
                                 zip(self.VRBD[idxs], self.Vcell[idxs],
                                     self.Icell[idxs])]
                            )
                            Icol, Vcol = self.pvconst.calcSeries(
                                self.Icell[idxs], self.Vcell[idxs],
                                self.Isc[idxs].mean(), IatVrbd.max()
                            )
                        else:
                            Icol, Vcol = self.Icell[idxs], self.Vcell[idxs]
                        IVcols.append([Icol, Vcol, self.Voc[idxs].sum()])
                    # append IVcols and continue
                    IVprev_cols.append(IVcols)
                    if prev_col:
                        # if circuits are same in both columns then continue
                        if not all(icol['crosstie'] == jcol['crosstie']
                                   for icol, jcol in zip(prev_col, col)):
                            # combine crosstied circuits
                            Iparallel, Vparallel = combine_parallel_circuits(
                                IVprev_cols, self.pvconst
                            )
                            IVall_cols.append([Iparallel, Vparallel])
                            # reset prev_col
                            prev_col = None
                            IVprev_cols = []
                            continue
                    # set prev_col and continue
                    prev_col = col
                # combine any remaining crosstied circuits in substring
                if not IVall_cols:
                    # combine crosstied circuits
                    Isub, Vsub = combine_parallel_circuits(
                        IVprev_cols, self.pvconst
                    )
                else:
                    Iparallel, Vparallel = zip(*IVall_cols)
                    Iparallel = np.asarray(Iparallel)
                    Vparallel = np.asarray(Vparallel)
                    Voc_parallel = np.asarray([
                        np.interp(np.float64(0), np.flipud(i_par), np.flipud(v_par))
                        for i_par, v_par in zip(Iparallel, Vparallel)])
                    Isub, Vsub = self.pvconst.calcParallel(
                        Iparallel, Vparallel, Vparallel.max(), Vparallel.min(),
                        Voc=Voc_parallel.mean()
                    )

            if self.Vbypass_config == DEFAULT_BYPASS:
                bypassed = Vsub < self.Vbypass
                Vsub[bypassed] = self.Vbypass
            elif self.Vbypass_config == CUSTOM_SUBSTR_BYPASS:
                if self.Vbypass[substr_idx] is None:
                    # no bypass for this substring
                    pass
                else:
                    # bypass the substring
                    bypassed = Vsub < self.Vbypass[substr_idx]
                    Vsub[bypassed] = self.Vbypass[substr_idx]
            elif self.Vbypass_config == MODULE_BYPASS:
                # module bypass value will be assigned after the for loop for substrings is over
                pass

            Isubstr.append(Isub)
            Vsubstr.append(Vsub)
            Isc_substr.append(np.interp(np.float64(0), Vsub, Isub))
            Imax_substr.append(Isub.max())

        Isubstr, Vsubstr = np.asarray(Isubstr), np.asarray(Vsubstr)
        Isc_substr = np.asarray(Isc_substr)
        Imax_substr = np.asarray(Imax_substr)
        Imod, Vmod = self.pvconst.calcSeries(
            Isubstr, Vsubstr, Isc_substr.mean(), Imax_substr.max()
        )

        # if entire module has only one bypass diode
        if self.Vbypass_config == MODULE_BYPASS:
            bypassed = Vmod < self.Vbypass[0]
            Vmod[bypassed] = self.Vbypass[0]
        else:
            pass

        Pmod = Imod * Vmod
        return Imod, Vmod, Pmod, Isubstr, Vsubstr

    def plotCell(self):
        """
        Plot cell I-V curves.
        Returns cellPlot : matplotlib.pyplot figure
        """
        cellPlot = plt.figure()
        plt.subplot(2, 2, 1)
        plt.plot(self.Vcell.T, self.Icell.T)
        plt.title('Cell Reverse I-V Characteristics')
        plt.ylabel('Cell Current, I [A]')
        plt.xlim(self.VRBD.min() - 1, 0)
        plt.ylim(0, self.Isc.mean() + 10)
        plt.grid()
        plt.subplot(2, 2, 2)
        plt.plot(self.Vcell.T, self.Icell.T)
        plt.title('Cell Forward I-V Characteristics')
        plt.ylabel('Cell Current, I [A]')
        plt.xlim(0, self.Voc.max())
        plt.ylim(0, self.Isc.mean() + 1)
        plt.grid()
        plt.subplot(2, 2, 3)
        plt.plot(self.Vcell.T, self.Pcell.T)
        plt.title('Cell Reverse P-V Characteristics')
        plt.xlabel('Cell Voltage, V [V]')
        plt.ylabel('Cell Power, P [W]')
        plt.xlim(self.VRBD.min() - 1, 0)
        plt.ylim((self.Isc.mean() + 10) * (self.VRBD.min() - 1), -1)
        plt.grid()
        plt.subplot(2, 2, 4)
        plt.plot(self.Vcell.T, self.Pcell.T)
        plt.title('Cell Forward P-V Characteristics')
        plt.xlabel('Cell Voltage, V [V]')
        plt.ylabel('Cell Power, P [W]')
        plt.xlim(0, self.Voc.max())
        plt.ylim(0, (self.Isc.mean() + 1) * self.Voc.max())
        plt.grid()
        plt.tight_layout()
        return cellPlot

    def plotMod(self):
        """
        Plot module I-V curves.
        Returns modPlot : matplotlib.pyplot figure
        """
        modPlot = plt.figure()
        ax = plt.subplot(2, 1, 1)
        plt.plot(self.Vmod, self.Imod)
        plt.title('Module I-V Characteristics')
        plt.ylabel('Module Current, I [A]')
        plt.ylim(ymin=0)
        plt.xlim(self.Vmod.min() - 1, self.Vmod.max() + 1)
        plt.grid()
        plt.subplot(2, 1, 2, sharex=ax)
        plt.plot(self.Vmod, self.Pmod)
        plt.title('Module P-V Characteristics')
        plt.xlabel('Module Voltage, V [V]')
        plt.ylabel('Module Power, P [W]')
        plt.ylim(ymin=0)
        plt.xlim(self.Vmod.min() - 1, self.Vmod.max() + 1)
        plt.grid()
        plt.tight_layout()
        return modPlot



# -*- coding: utf-8 -*-
"""
This module contains the :class:`~pvmismatch.pvmismatch_lib.pvsystem.PVsystem`
class.
"""

from __future__ import absolute_import
from past.builtins import basestring
from future.utils import iteritems
import numpy as np
from copy import copy
from matplotlib import pyplot as plt
# use absolute imports instead of relative, so modules are portable
from .pvconstants import PVconstants, NUMBERMODS, NUMBERSTRS
from .pvstring import PVstring


class PVsystem(object):
    """
    A class for PV systems.

    :param pvconst: configuration constants object
    :type pvconst: :class:`~pvmismatch.pvmismatch_lib.pvconstants.PVconstants`
    :param numberStrs: number of strings
    :param pvstrs: list of parallel strings, a ``PVstring`` object or None
    :param numberMods: number of modules per string
    :param pvmods: list of modules, a ``PVmodule`` object or None
    """

    def __init__(self, pvconst=None, numberStrs=NUMBERSTRS,
                 pvstrs=None, numberMods=NUMBERMODS, pvmods=None):
        # is pvstrs a list?
        try:
            pvstr0 = pvstrs[0]
        except TypeError:
            # is pvstrs a PVstring object?
            try:
                pvconst = pvstrs.pvconst
            except AttributeError:
                # try to use the pvconst arg or create one if none
                if not pvconst:
                    pvconst = PVconstants()
                # create a pvstring
                pvstrs = PVstring(numberMods=numberMods, pvmods=pvmods,
                                  pvconst=pvconst)
            # expand pvstrs to list
            pvstrs = [pvstrs] * numberStrs
            numberMods = [numberMods] * numberStrs
        else:
            pvconst = pvstr0.pvconst
            numberStrs = len(pvstrs)
            numberMods = []
            for p in pvstrs:
                if p.pvconst is not pvconst:
                    raise Exception('pvconst must be the same for all strings')
                numberMods.append(len(p.pvmods))
        self.pvconst = pvconst  #: ``PVconstants`` used in ``PVsystem``
        self.numberStrs = numberStrs  #: number strings in the system
        self.numberMods = numberMods  #: list of number of modules per string
        self.pvstrs = pvstrs  #: list of ``PVstring`` in system
        # calculate pvsystem
        self.update()

    def update(self):
        """Update system calculations."""
        self.Isys, self.Vsys, self.Psys = self.calcSystem()
        (self.Imp, self.Vmp, self.Pmp,
         self.Isc, self.Voc, self.FF, self.eff) = self.calcMPP_IscVocFFeff()

    # TODO: use __getattr__ to check for updates to pvcells

    @property
    def pvmods(self):
        return [pvstr.pvmods for pvstr in self.pvstrs]

    @property
    def Istring(self):
        return np.asarray([pvstr.Istring.flatten() for pvstr in self.pvstrs])

    @property
    def Vstring(self):
        return np.asarray([pvstr.Vstring.flatten() for pvstr in self.pvstrs])

    @property
    def Voc_str(self):
        return np.asarray([pvstr.Voc_mod.sum() for pvstr in self.pvstrs])

    def calcSystem(self):
        """
        Calculate system I-V curves.
        Returns (Isys, Vsys, Psys) : tuple of numpy.ndarray of float
        """
        Isys, Vsys = self.pvconst.calcParallel(
            self.Istring, self.Vstring, self.Voc_str.max(), self.Vstring.min()
        )
        Psys = Isys * Vsys
        return Isys, Vsys, Psys

    def calcMPP_IscVocFFeff(self):
        mpp = np.argmax(self.Psys)
        P = self.Psys[mpp - 1:mpp + 2]
        V = self.Vsys[mpp - 1:mpp + 2]
        I = self.Isys[mpp - 1:mpp + 2]
        # calculate derivative dP/dV using central difference
        dP = np.diff(P, axis=0)  # size is (2, 1)
        dV = np.diff(V, axis=0)  # size is (2, 1)
        Pv = dP / dV  # size is (2, 1)
        # dP/dV is central difference at midpoints,
        Vmid = (V[1:] + V[:-1]) / 2.0  # size is (2, 1)
        Imid = (I[1:] + I[:-1]) / 2.0  # size is (2, 1)
        # interpolate to find Vmp
        Vmp = (-Pv[0] * np.diff(Vmid, axis=0) / np.diff(Pv, axis=0) + Vmid[0]).item()
        Imp = (-Pv[0] * np.diff(Imid, axis=0) / np.diff(Pv, axis=0) + Imid[0]).item()
        # calculate max power at Pv = 0
        Pmp = Imp * Vmp
        # calculate Voc, current must be increasing so flipup()
        Voc = np.interp(np.float64(0), np.flipud(self.Isys),
                        np.flipud(self.Vsys))
        Isc = np.interp(np.float64(0), self.Vsys, self.Isys)  # calculate Isc
        FF = Pmp / Isc / Voc
        totalSuns = sum(
            [pvmod.Ee.sum() * pvmod.cellArea for pvstr in self.pvmods
             for pvmod in pvstr]
        )
        # convert cellArea from cm^2 to m^2
        Psun = self.pvconst.E0 * totalSuns / 100 / 100
        eff = Pmp / Psun
        return Imp, Vmp, Pmp, Isc, Voc, FF, eff

    def setSuns(self, Ee):
        """
        Set irradiance on cells in modules of string in system.
        If Ee is ...
        ... scalar, then sets the entire system to that irradiance.
        ... a dictionary, then each key refers to a pv-string in the system,
        and the corresponding value is either a dictionary or a scalar.
        If the dictionary's value is another dictionary, then its keys are pv-
        modules and its values are either cells and corresponding Ee, cells and
        a scalar Ee, a scalar Ee value for all cells or an array of Ee values
        for all cells in the module. The values of pv-modules are passed to
        :meth:`~pvmismatch.pvmismatch_lib.pvmodule.PVmodule.setSuns()`

        :param Ee: irradiance [suns]
        :type Ee: dict, float

        For Example::

            Ee={0: {0: {'cells': (0, 1, 2), 'Ee': (0.9, 0.3, 0.5)}}}
            Ee=0.91  # set all modules in all strings to 0.91 suns
            Ee={12: 0.77}  # set all modules in string with index 12 to 0.77 suns
            Ee={3: {8: 0.23, 7: 0.45}}
            # set module with index 8 to 0.23 suns and module with index 7 to
            # 0.45 suns in string with index 3

        """
        if np.isscalar(Ee):
            for pvstr in self.pvstrs:
                pvstr.setSuns(Ee)
        else:
            for pvstr, pvmod_Ee in iteritems(Ee):
                pvstr = int(pvstr)
                self.pvstrs[pvstr] = copy(self.pvstrs[pvstr])
                self.pvstrs[pvstr].setSuns(pvmod_Ee)
        # calculate pvsystem
        self.update()

    def setTemps(self, Tc):
        """
        Set temperature on cells in modules of string in system.
        If Tc is ...
        ... scalar, then sets the entire system to that cell temperature.
        ... a dictionary, then each key refers to a pv-string in the system,
        and the corresponding value is either a dictionary or a scalar.
        If the dictionary's value is another dictionary, then its keys are pv-
        modules and its values are either cells and corresponding Tc, cells and
        a scalar Tc, a scalar Tc value for all cells or an array of Tc values
        for all cells in the module. The values of pv-modules are passed to
        :meth:`~pvmismatch.pvmismatch_lib.pvmodule.PVmodule.setTemps()`

        :param Tc: temperature [K]
        :type Tc: dict, float

        For Example::

            Tc={0: {0: {'cells': (1,2,3), 'Tc': (323.15, 348.15, 373.15)}}}
            Tc=323.15  # set all modules in all strings to 323.15K (50°C)
            Tc={12: 348.15}  # set all modules in string with index 12 to 348.15K (75°C)
            Tc={3: {8: 333.15, 7: 373.15}}
            # set module with index 8 to 333.15K (60°C) and module with index 7 to
            # 373.15K (100°C) in string with index 3

        """
        if np.isscalar(Tc):
            for pvstr in self.pvstrs:
                pvstr.setTemps(Tc)
        else:
            for pvstr, pvmod_Tc in iteritems(Tc):
                pvstr = int(pvstr)
                self.pvstrs[pvstr] = copy(self.pvstrs[pvstr])
                self.pvstrs[pvstr].setTemps(pvmod_Tc)
        # calculate pvsystem
        self.update()

    def plotSys(self, sysPlot=None, fmt=''):
        """
        Plot system I-V curves.

        :param sysPlot: integer, string, or existing figure
        :returns: new figure
        """
        # create new figure if sysPlot or make the specified sysPlot current
        # and clear it
        try:
            sysPlot.clear()
        except (AttributeError, SyntaxError):
            sysPlot = plt.figure(sysPlot)
        ax = plt.subplot(2, 1, 1)
        plt.plot(self.Vsys, self.Isys, fmt)
        plt.xlim(0, self.Voc * 1.1)
        plt.ylim(0, self.Isc * 1.1)
        plt.axvline(self.Vmp, color='r', linestyle=':')
        plt.axhline(self.Imp, color='r', linestyle=':')
        plt.title('System I-V Characteristics')
        plt.ylabel('System Current, I [A]')
        plt.grid()
        plt.subplot(2, 1, 2, sharex=ax)
        plt.plot(self.Vsys, self.Psys / 1000, fmt)
        plt.xlim(0, self.Voc * 1.1)
        plt.ylim(0, self.Pmp * 1.1 / 1000)
        plt.axvline(self.Vmp, color='r', linestyle=':')
        plt.axhline(self.Pmp / 1000, color='r', linestyle=':')
        plt.title('System P-V Characteristics')
        plt.xlabel('System Voltage, V [V]')
        plt.ylabel('System Power, P [kW]')
        plt.grid()
        plt.tight_layout()
        return sysPlot



# -*- coding: utf-8 -*-
"""
Subfolder containing the PVMismatch Library.

Configuration constants and classes like
:class:`~pvmismatch.pvmismatch_lib.pvconstants.PVconstants`,
:class:`~pvmismatch.pvmismatch_lib.pvcell.PVcell`,
:class:`~pvmismatch.pvmismatch_lib.pvmodule.PVmodule`,
:class:`~pvmismatch.pvmismatch_lib.pvstring.PVstring` and
:class:`~pvmismatch.pvmismatch_lib.pvsystem.PVsystem`, objects are all defined
here.
"""

from v_PVMismatch.pvmismatch import pvconstants, pvexceptions, pvcell, pvmodule, pvstring, pvsystem



# -*- coding: utf-8 -*-
"""
Initialization file for the vectorized PVMismatch (v_PVMismatch) Package.

This package contains the basic library modules, methods, classes and
attributes to model PV system mismatch.
    >>> from v_pvmismatch import pvcell  # imports the PVcell methods
    >>> # import pvcell, pvmodule, pvstring and pvsystem
    >>> from v_pvmismatch import *
"""
# from v_pvmismatch.version import __version__

from v_PVMismatch import (
    utils,
    circuit_comb,
    vpvcell,
    vpvmodule,
    vpvstring,
    vpvsystem,
    pvmismatch
)



# -*- coding: utf-8 -*-
"""Helper functions that are vectorized."""

import numpy as np
from functools import partial
from scipy.interpolate import interp1d

# ------------------------------------------------------------------------------
# HELPER FUNCTIONS THAT ARE VECTORIZED ----------------------------------------


def reshape_ndarray(x, arr_shape):
    """
    Reshape an N dimensional array given the new array shape.

    Parameters
    ----------
    x : numpy.ndarray
        Array of N-dimensions.
    arr_shape : tuple
        Tuple specifying the shape of the input array.

    Returns
    -------
    xout : numpy.ndarray
        Output array reshaped to input shape.

    """
    xout = x.copy()
    num_new_axis = len(arr_shape)
    # Run for loop over the new axis to create
    for idx_ax in range(num_new_axis):
        xout = np.repeat(xout[..., np.newaxis],
                         arr_shape[idx_ax], axis=idx_ax+1)

    return xout


def view1D(a, b):
    """
    Create 1d views of a, b.

    Parameters
    ----------
    a : numpy.ndarray
        Array of N-dimensions.
    b : numpy.ndarray
        Array of N-dimensions.

    Returns
    -------
    numpy.ndarray
        View of a.
    numpy.ndarray
        View of b.

    """
    a = np.ascontiguousarray(a)
    b = np.ascontiguousarray(b)
    void_dt = np.dtype((np.void, a.dtype.itemsize * a.shape[1]))
    return a.view(void_dt).ravel(),  b.view(void_dt).ravel()


def isin_nd(a, b):
    """
    Reproduce isin in N-dimensions.

    Parameters
    ----------
    a : numpy.ndarray
        3D input array.
    b : numpy.ndarray
        3D input array.

    Returns
    -------
    numpy.ndarray
        DESCRIPTION.

    """
    # a,b are the 3D input arrays to give us "isin-like" functionality across them
    A, B = view1D(a.reshape(a.shape[0], -1), b.reshape(b.shape[0], -1))
    return np.nonzero(np.isin(A, B))[0]


def isin_nd_searchsorted(a, b):
    """
    Perform isin and search sorting in N-dimensions.

    Parameters
    ----------
    a : numpy.ndarray
        3D input array.
    b : numpy.ndarray
        3D input array.

    Returns
    -------
    numpy.ndarray
        Boolean array.

    """
    # a,b are the 3D input arrays
    A, B = view1D(a.reshape(a.shape[0], -1), b.reshape(b.shape[0], -1))
    sidx = A.argsort()
    sorted_index = np.searchsorted(A, B, sorter=sidx)
    sorted_index[sorted_index == len(A)] = len(A)-1
    idx = sidx[sorted_index]
    return A[idx] == B


# def searchsorted2d(a, b):
#     # https://stackoverflow.com/questions/40588403/vectorized-searchsorted-numpy
#     m, n = a.shape
#     max_num = np.maximum(a.max() - a.min(), b.max() - b.min()) + 1
#     r = max_num*np.arange(a.shape[0])[:, None]
#     p = np.searchsorted((a+r).ravel(), (b+r).ravel(),
#                         side='right').reshape(m, -1)
#     return p - n*(np.arange(m)[:, None])


def interp_coef(x0: np.ndarray, x: np.ndarray, side='right', kind='linear'):
    """
    Calculate interpolation coefficiencts for N-dimensional arrays.

    Parameters
    ----------
    x0 : numpy.ndarray
        Array to interpolate from.
    x : numpy.ndarray
        Array to interpolate to.
    side : str, optional
        Which side to interpolate from (input for numpy.searchsorted). The default is 'right'.
    kind : str, optional
        What type of interpolation. The options are 'linear', 'previous', and 'next'. The default is 'linear'.

    Returns
    -------
    lo_col : numpy.ndarray
        Left index.
    hi_col : numpy.ndarray
        Right index.
    hi_row : numpy.ndarray
        Rows.
    w : numpy.ndarray
        Weights for interpolation.

    """
    # https://towardsdatascience.com/linear-interpolation-in-python-a-single-line-of-code-25ab83b764f9
    # Modified to do sorting in 2d
    # find the indices into the original array
    # hi_col = np.minimum(x0.shape[1] - 1, searchsorted2d(x0, x)) # Doesn't work for large float numbers >1e20
    mapfunc = partial(np.searchsorted, side=side)
    hi_col = np.minimum(
        x0.shape[1] - 1, np.array(list(map(mapfunc, x0, x))))
    hi_row = np.array(list(range(hi_col.shape[0])))
    hi_row = np.repeat(hi_row, hi_col.shape[1])
    lo_col = np.maximum(0, hi_col - 1)

    # calculate the distance within the range
    if kind == 'linear':
        d_left = x - x0[hi_row.flatten(), lo_col.flatten()].reshape(lo_col.shape)
        d_right = x0[hi_row.flatten(), hi_col.flatten()].reshape(hi_col.shape) - x
        d_total = d_left + d_right
        # weights are the proportional distance
        w = d_right / d_total
    elif kind == 'previous':
        w = np.ones(lo_col.shape)
    elif kind == 'next':
        w = np.zeros(lo_col.shape)
    # correction if we're outside the range of the array
    # w[np.isinf(w)] = 0.0
    # return the information contained by the projection matrices
    return (lo_col, hi_col, hi_row, w)


def interp_2d(y0: np.ndarray, coef, x, x0):
    """
    Perform interpolation in N-dimensions with base x0.

    Parameters
    ----------
    y0 : np.ndarray
        Y Array to interpolate from.
    coef : tuple
        Output tuple from interp_coef.
    x : np.ndarray
        X Array to interpolate to.
    x0 : np.ndarray
        X Array to interpolate from.

    Returns
    -------
    yint : np.ndarray
        Interpolated Y array.

    """
    # https://towardsdatascience.com/linear-interpolation-in-python-a-single-line-of-code-25ab83b764f9
    term1 = coef[3] * y0[coef[2].flatten(), coef[0].flatten()
                         ].reshape(coef[0].shape)
    term2 = (1 - coef[3]) * y0[coef[2].flatten(),
                               coef[1].flatten()].reshape(coef[0].shape)
    yint = term1 + term2
    # Extrapolate
    # extrapolate left only
    x0_fr = x0[:, 0]
    left = np.logical_and(np.isnan(yint), x < x0_fr[:, None])
    idx_nan = np.where(left)
    xleft = x[left]
    yleft = y0[idx_nan[0], 0] + (xleft - x0[idx_nan[0], 0]) / \
        (x0[idx_nan[0], 1] - x0[idx_nan[0], 0]) * (y0[idx_nan[0], 1] - y0[idx_nan[0], 0])
    yint[idx_nan[0], idx_nan[1]] = yleft
    # extrapolate right only
    # extrapolate right
    x0_lr = x0[:, -1]
    right = np.logical_and(np.isnan(yint), x > x0_lr[:, None])
    idx_nan = np.where(right)
    xright = x[right]
    yright = y0[idx_nan[0], -1] + (xright - x0[idx_nan[0], -1]) / \
        (x0[idx_nan[0], -2] - x0[idx_nan[0], -1]) * (y0[idx_nan[0], -2] - y0[idx_nan[0], -1])
    yint[idx_nan[0], idx_nan[1]] = yright
    return yint


def interp2d_wrap(x0, x, y0, side='right', kind='linear'):
    """
    Wrap function for N-dimensional interpolation.

    Parameters
    ----------
    x0 : np.ndarray
        X Array to interpolate from.
    x : np.ndarray
        X Array to interpolate to.
    y0 : np.ndarray
        Y Array to interpolate from.
    side : str, optional
        Which side to interpolate from (input for numpy.searchsorted). The default is 'right'.
    kind : str, optional
        What type of interpolation. The options are 'linear', 'previous', and 'next'. The default is 'linear'.

    Returns
    -------
    yint : np.ndarray
        Interpolated Y array.

    """
    # Calculate weighting coefficients
    coef = interp_coef(x0, x, side=side, kind=kind)
    # Interpolate
    yint = interp_2d(y0, coef, x, x0)

    return yint


def calcMPP_IscVocFFBPD(Isys, Vsys, Psys, bypassed_mod_arr, run_bpact=True, run_annual=False):
    """
    Calculate MPP IV parameters. This method is vectorized.

    Parameters
    ----------
    Isys : np.ndarray
        2-D array of current curves.
    Vsys : np.ndarray
        2-D array of voltage curves.
    Psys : np.ndarray
        2-D array of power curves.
    bypassed_mod_arr : np.ndarray
        3-D or 5-D array of bypass diode activation curves for each substring in each module of the system.
    run_bpact : bool, optional
        Flag to run bypass diode activation logic. The default is True.
    run_annual : bool, optional
        Flag to delete large bypass diode activation array in case of an annual simulation. The default is False.

    Returns
    -------
    Imp : np.ndarray
        1-D array of Imp.
    Vmp : np.ndarray
        1-D array of Vmp.
    Pmp : np.ndarray
        1-D array of Pmp.
    Isc : np.ndarray
        1-D array of Isc.
    Voc : np.ndarray
        1-D array of Voc.
    FF : np.ndarray
        1-D array of FF.
    BpDmp : np.ndarray
        Bypass diode activation at MPP for each substring in each module of the system.
    num_bpd_active : np.ndarray
        1-D array of number of bypass diodes active.

    """
    # Reverse direction of Psys
    rev_P = Psys[:, ::-1]
    mpp = rev_P.shape[1] - np.argmax(rev_P, axis=1) - 1
    check_max_idx = (mpp == rev_P.shape[1]-1)
    mpp[check_max_idx] = rev_P.shape[1] - 2
    mpp = np.reshape(mpp, (len(mpp), 1))
    mpp_lohi = np.concatenate([mpp-1, mpp, mpp+1], axis=1)
    mpp_row = np.reshape(np.arange(Psys.shape[0]), (Psys.shape[0], 1))
    P = Psys[mpp_row, mpp_lohi]
    V = Vsys[mpp_row, mpp_lohi]
    Curr = Isys[mpp_row, mpp_lohi]
    # calculate derivative dP/dV using central difference
    dP = np.diff(P, axis=1)  # size is (2, 1)
    dV = np.diff(V, axis=1)  # size is (2, 1)
    Pv = dP / dV  # size is (2, 1)
    # dP/dV is central difference at midpoints,
    Vmid = (V[:, 1:] + V[:, :-1]) / 2.0  # size is (2, 1)
    Imid = (Curr[:, 1:] + Curr[:, :-1]) / 2.0  # size is (2, 1)
    # interpolate to find Vmp
    Vmp = (-Pv[:, 0].flatten() * np.diff(Vmid, axis=1).flatten() /
           np.diff(Pv, axis=1).flatten() + Vmid[:, 0])
    Imp = (-Pv[:, 0].flatten() * np.diff(Imid, axis=1).flatten() /
           np.diff(Pv, axis=1).flatten() + Imid[:, 0])
    # calculate max power at Pv = 0
    Pmp = Imp * Vmp
    # calculate Voc, current must be increasing so flipup()
    Voc = np.zeros(Pmp.shape)
    Isc = np.zeros(Pmp.shape)
    for idx_time in range(Psys.shape[0]):
        # Only interpolate if Current data is non-zero
        if Vsys[idx_time, :].nonzero()[0].size != 0:
            Voc[idx_time] = np.interp(np.float64(0), np.flipud(Isys[idx_time, :]),
                                      np.flipud(Vsys[idx_time, :]))
            Isc[idx_time] = np.interp(np.float64(
                0), Vsys[idx_time, :], Isys[idx_time, :])  # calculate Isc
    FF = Pmp / Isc / Voc
    if run_bpact:
        # Use nearest interpolation to obtain bypass diode activation at MPP
        if len(bypassed_mod_arr.shape) == 3:
            BpD_Active = bypassed_mod_arr[mpp_row, :, mpp_lohi]
            BpDmp = np.zeros((Pmp.shape[0], BpD_Active.shape[2]), dtype=bool)
            for idx_row in range(BpD_Active.shape[0]):
                for idx_col in range(BpD_Active.shape[2]):
                    interpolator = interp1d(P[idx_row, :], BpD_Active[idx_row, :, idx_col], kind='previous',
                                            fill_value='extrapolate')
                    BpDmp[idx_row, idx_col] = interpolator(Pmp[idx_row]).astype(bool)
            num_bpd_active = BpDmp.sum(axis=1)
        else:
            BpD_Active = bypassed_mod_arr[mpp_row, :, :, :, mpp_lohi]
            BpDmp = np.zeros((Pmp.shape[0], BpD_Active.shape[2], BpD_Active.shape[3], BpD_Active.shape[4]), dtype=bool)
            for idx_row in range(BpD_Active.shape[0]):
                for idx_str in range(BpD_Active.shape[2]):
                    for idx_mod in range(BpD_Active.shape[3]):
                        for idx_substr in range(BpD_Active.shape[4]):
                            interpolator = interp1d(P[idx_row, :],
                                                    BpD_Active[idx_row, :, idx_str, idx_mod,
                                                               idx_substr], kind='previous',
                                                    fill_value='extrapolate')
                            BpDmp[idx_row, idx_str, idx_mod, idx_substr] = interpolator(Pmp[idx_row]).astype(bool)
            num_bpd_active = BpDmp.sum(axis=3).sum(axis=2).sum(axis=1)
    else:
        BpDmp = np.nan
        num_bpd_active = 0
    if run_annual:
        del bypassed_mod_arr

    return Imp, Vmp, Pmp, Isc, Voc, FF, BpDmp, num_bpd_active



# -*- coding: utf-8 -*-
"""
This module contains the :class:`~pvmismatch.pvmismatch_lib.pvstring.PVstring`
class.
"""

from __future__ import absolute_import
from past.builtins import range
from future.utils import iteritems
import numpy as np
from copy import copy
from matplotlib import pyplot as plt
# use absolute imports instead of relative, so modules are portable
from .pvconstants import PVconstants, NUMBERMODS
from .pvmodule import PVmodule


class PVstring(object):
    """
    A class for PV strings.

    :param numberMods: number of modules in string
    :param pvmods: either a list of modules, an instance of
        :class:`~pvmismatch.pvmismatch_lib.pvmodule.PVmodule` [Optional]
    :type pvmods: list, :class:`~pvmismatch.pvmismatch_lib.pvmodule.PVmodule`
    :param pvconst: a configuration constants object
    """

    def __init__(self, numberMods=NUMBERMODS, pvmods=None,
                 pvconst=None):
        # is pvmods a list?
        try:
            pvmod0 = pvmods[0]
        except TypeError:
            # is pvmods an object?
            try:
                pvconst = pvmods.pvconst
            except AttributeError:
                #  try to use the pvconst arg or create one if none
                if not pvconst:
                    pvconst = PVconstants()
                # create pvmod
                pvmods = PVmodule(pvconst=pvconst)
            # expand pvmods to list
            pvmods = [pvmods] * numberMods
        else:
            pvconst = pvmod0.pvconst
            numberMods = len(pvmods)
            for p in pvmods:
                if p.pvconst is not pvconst:
                    raise Exception('pvconst must be the same for all modules')
        self.pvconst = pvconst  #: ``PVconstants`` used in  ``PVstring``
        self.numberMods = numberMods  #: number of module in string
        self.pvmods = pvmods  #: list of ``PVModule`` in ``PVstring``
        # calculate string
        self.Istring, self.Vstring, self.Pstring = self.calcString()

    # TODO: use __getattr__ to check for updates to pvcells

    @property
    def Imod(self):
        return np.array([mod.Imod.flatten() for mod in self.pvmods])

    @property
    def Vmod(self):
        return np.array([mod.Vmod.flatten() for mod in self.pvmods])

    @property
    def Voc_mod(self):
        return np.array([mod.Voc.sum() for mod in self.pvmods])

    def calcString(self):
        """
        Calculate string I-V curves.
        Returns (Istring, Vstring, Pstring) : tuple of numpy.ndarray of float
        """
        # Imod is already set to the range from Vrbd to the minimum current
        meanIsc = np.mean([pvmod.Isc.mean() for pvmod in self.pvmods])
        Istring, Vstring = self.pvconst.calcSeries(self.Imod, self.Vmod,
                                                   meanIsc, self.Imod.max())
        Pstring = Istring * Vstring
        return Istring, Vstring, Pstring

    def setSuns(self, Ee):
        """
        Set irradiance on cells in modules of string in system.
        If Ee is ...
        ... scalar, then sets the entire string to that irradiance.
        ... a dictionary, then each key refers to a module in the string,
        and the corresponding value are passed to
        :meth:`~pvmismatch.pvmismatch_lib.pvmodule.PVmodule.setSuns()`

        :param Ee: irradiance [suns]
        :type Ee: dict or float

        For Example::

            Ee={0: {'cells': (1,2,3), 'Ee': (0.9, 0.3, 0.5)}}  # set module 0
            Ee=0.91  # set all modules to 0.91 suns
            Ee={12: 0.77}  # set module with index 12 to 0.77 suns
            Ee={8: [0.23, (0, 1, 2)], 7: [(0.45, 0.35), (71, 72)]}
            # set module 8, cells 0, 1 and 2 to 0.23 suns, then set module 7, cell
            #   71 to 0.45 suns and module 72 to 0.35 suns.

        """
        if np.isscalar(Ee):
            new_pvmods = range(self.numberMods)  # new list of modules
            old_pvmods = dict.fromkeys(self.pvmods)  # same as set(pvmods)
            for mod_id, pvmod in enumerate(self.pvmods):
                if old_pvmods[pvmod] is None:
                    new_pvmods[mod_id] = copy(pvmod)
                    old_pvmods[pvmod] = new_pvmods[mod_id]
                else:
                    new_pvmods[mod_id] = old_pvmods[pvmod]
            self.pvmods = new_pvmods
            for pvmod in iter(self.pvmods):
                pvmod.setSuns(Ee)
        else:
            self.pvmods = copy(self.pvmods)  # copy list first
            try:
                for pvmod, cell_Ee in iteritems(Ee):
                    pvmod = int(pvmod)
                    self.pvmods[pvmod] = copy(self.pvmods[pvmod])
                    if hasattr(cell_Ee, 'keys'):
                        self.pvmods[pvmod].setSuns(**cell_Ee)
                    else:
                        try:
                            self.pvmods[pvmod].setSuns(*cell_Ee)
                        except TypeError:
                            self.pvmods[pvmod].setSuns(cell_Ee)
            except AttributeError:
                # Ee was a list? just take first item in list
                if len(Ee) > 1:
                    raise TypeError('Irradiance, Ee, should be scalar or dict')
                Ee = Ee[0]
                new_pvmods = range(self.numberMods)  # new list of modules
                old_pvmods = dict.fromkeys(self.pvmods)  # same as set(pvmods)
                for mod_id, pvmod in enumerate(self.pvmods):
                    if old_pvmods[pvmod] is None:
                        new_pvmods[mod_id] = copy(pvmod)
                        old_pvmods[pvmod] = new_pvmods[mod_id]
                    else:
                        new_pvmods[mod_id] = old_pvmods[pvmod]
                self.pvmods = new_pvmods
                for pvmod in iter(self.pvmods):
                    pvmod.setSuns(Ee)
        # update modules
        self.Istring, self.Vstring, self.Pstring = self.calcString()

    def setTemps(self, Tc):
        """
        Set temperature on cells in modules of string in system.
        If Tc is ...
        ... scalar, then sets the entire string to that temperature.
        ... a dictionary, then each key refers to a module in the string,
        and the corresponding value are passed to
        :meth:`~pvmismatch.pvmismatch_lib.pvmodule.PVmodule.setTemps()`

        :param Tc: temperature [K]
        :type Tc: dict or float

        For Example::

            Tc={0: {'cells': (1,2,3), 'Tc': (323.15, 348.15, 373.15)}}  # set module 0
            Tc=323.15  # set all modules to 323.15K (50°C)
            Tc={12: 348.15}  # set module with index 12 to 348.15K (75°C)

        """
        if np.isscalar(Tc):
            new_pvmods = range(self.numberMods)  # new list of modules
            old_pvmods = dict.fromkeys(self.pvmods)  # same as set(pvmods)
            for mod_id, pvmod in enumerate(self.pvmods):
                if old_pvmods[pvmod] is None:
                    new_pvmods[mod_id] = copy(pvmod)
                    old_pvmods[pvmod] = new_pvmods[mod_id]
                else:
                    new_pvmods[mod_id] = old_pvmods[pvmod]
            self.pvmods = new_pvmods
            for pvmod in iter(self.pvmods):
                pvmod.setTemps(Tc)
        else:
            self.pvmods = copy(self.pvmods)  # copy list first
            try:
                for pvmod, cell_Tc in iteritems(Tc):
                    pvmod = int(pvmod)
                    self.pvmods[pvmod] = copy(self.pvmods[pvmod])
                    if hasattr(cell_Tc, 'keys'):
                        self.pvmods[pvmod].setTemps(**cell_Tc)
                    else:
                        try:
                            self.pvmods[pvmod].setTemps(*cell_Tc)
                        except TypeError:
                            self.pvmods[pvmod].setTemps(cell_Tc)
            except AttributeError:
                # Tc was a list? just take first item in list
                if len(Tc) > 1:
                    raise TypeError('Irradiance, Ee, should be scalar or dict')
                Tc = Tc[0]
                new_pvmods = range(self.numberMods)  # new list of modules
                old_pvmods = dict.fromkeys(self.pvmods)  # same as set(pvmods)
                for mod_id, pvmod in enumerate(self.pvmods):
                    if old_pvmods[pvmod] is None:
                        new_pvmods[mod_id] = copy(pvmod)
                        old_pvmods[pvmod] = new_pvmods[mod_id]
                    else:
                        new_pvmods[mod_id] = old_pvmods[pvmod]
                self.pvmods = new_pvmods
                for pvmod in iter(self.pvmods):
                    pvmod.setTemps(Tc)
        # update modules
        self.Istring, self.Vstring, self.Pstring = self.calcString()

    def plotStr(self):
        """
        Plot string I-V curves.
        Returns strPlot : matplotlib.pyplot figure
        """
        strPlot = plt.figure()
        ax = plt.subplot(2, 1, 1)
        plt.plot(self.Vstring, self.Istring)
        plt.title('String I-V Characteristics')
        plt.ylabel('String Current, I [A]')
        plt.ylim(ymin=0)
        plt.grid()
        plt.subplot(2, 1, 2, sharex=ax)
        plt.plot(self.Vstring, self.Pstring)
        plt.title('String P-V Characteristics')
        plt.xlabel('String Voltage, V [V]')
        plt.ylabel('String Power, P [W]')
        plt.grid()
        plt.tight_layout()
        return strPlot



# -*- coding: utf-8 -*-
"""Vectorized pvsystem."""

import numpy as np
from .utils import reshape_ndarray, isin_nd, calcMPP_IscVocFFBPD
from .vpvstring import calcStrings
from .circuit_comb import calcParallel_with_bypass

# ------------------------------------------------------------------------------
# BUILD SYSTEM IV-PV CURVES----------------------------------------------------


def gen_sys_Ee_Tcell_array(sim_len, Num_mod_X, Num_mod_Y, Num_cell_X, Num_cell_Y,
                           Ee=1., Tcell=298.15):
    """
    Generate system level irradiance and cell temperature arrays.

    Parameters
    ----------
    sim_len : int
        Length of the simulation or number of combinations.
    Num_mod_X : int
        Number of modules in a string.
    Num_mod_Y : int
        Number of strings in a system.
    Num_cell_X : int
        Number of columns in a module.
    Num_cell_Y : int
        Number of rows in a module.
    Ee : float or numpy.ndarray, optional
        Input irradiance value or array. The default is 1..
    Tcell : float or numpy.ndarray, optional
        Input cell temperature value or array. The default is 298.15.

    Returns
    -------
    Ee : numpy.ndarray
        System level irradiance array.
    Tcell : numpy.ndarray
        System level cell temperature array.

    """
    # Check if Ee is scalar
    if not (isinstance(Ee, np.ndarray)):
        Ee = Ee*np.ones((sim_len, Num_mod_X, Num_mod_Y,
                        Num_cell_X, Num_cell_Y))
    # Check if Ee is a 1d array with different Ee for each sim index
    elif len(Ee.shape) == 1:
        Ee = reshape_ndarray(
            Ee, (Num_mod_X, Num_mod_Y, Num_cell_X, Num_cell_Y))
    # Check if Tcell is scalar
    if not (isinstance(Tcell, np.ndarray)):
        Tcell = Tcell * \
            np.ones((sim_len, Num_mod_X, Num_mod_Y, Num_cell_X, Num_cell_Y))
    # Check if Tcell is a 1d array with different Ee for each sim index
    elif len(Tcell.shape) == 1:
        Tcell = reshape_ndarray(
            Tcell, (Num_mod_X, Num_mod_Y, Num_cell_X, Num_cell_Y))

    return Ee, Tcell


def get_unique_Ee(Ee, search_type='cell', cell_type=None):
    """
    Generate unique irradiance at the cell, module, string or system level.

    Parameters
    ----------
    Ee : numpy.ndarray
        System level irradiance array.
    search_type : str, optional
        Which type of unique irradiance: 'cell', 'module', 'string', or 'system'. The default is 'cell'.
    cell_type : numpy.ndarray or None, optional
        Array containing the different cell types. The default is None.

    Returns
    -------
    u : numpy.ndarray
        Unique irradiance values.
    u_cell_type : numpy.ndarray or None
        Unique cell types sorted based on u but only for cell level search.

    """
    # Get the unique Ee for cell level
    if search_type == 'cell':
        if isinstance(cell_type, np.ndarray):
            u_ctype = np.unique(cell_type)
            u = []
            u_cell_type = []
            for uct in u_ctype:
                idx_ct = np.where(cell_type == uct)
                Ee_sub = Ee[:, :, :, idx_ct[0], idx_ct[1]]
                usub = np.unique(Ee_sub, axis=None)
                ctype = uct * np.ones(usub.shape)
                u.append(usub)
                u_cell_type.append(ctype)
            u = np.concatenate(u)
            u_cell_type = np.concatenate(u_cell_type)
        else:
            u = np.unique(Ee, axis=None)
            u_cell_type = 0 * np.ones(u.shape)
    elif search_type == 'module':
        Ee_shp = Ee.shape
        u_list = []
        for idx_0 in range(Ee_shp[0]):
            for idx_1 in range(Ee_shp[1]):
                u_list.append(np.unique(Ee[idx_0, idx_1, :], axis=0))
        u = np.concatenate(u_list)
        u = np.unique(u, axis=0)
        u_cell_type = None
    elif search_type == 'string':
        Ee_shp = Ee.shape
        u_list = []
        for idx_0 in range(Ee_shp[0]):
            u_list.append(np.unique(Ee[idx_0, :], axis=0))
        u = np.concatenate(u_list)
        u = np.unique(u, axis=0)
        u_cell_type = None
    elif search_type == 'system':
        u = np.unique(Ee, axis=0)
        u_cell_type = None
    else:
        print('Incorrect search_type inputted. Allowed search_type: cell, module, string, system.')

    return u, u_cell_type


def calcTimeSeries(Ee_vec, sys_data):
    """
    Generate IV curves for the entire simulation or all combinations.

    Parameters
    ----------
    Ee_vec : numpy.ndarray
        System level irradiance array.
    sys_data : dict
        System level IV curves.

    Returns
    -------
    time_data : dict
        Simulation level IV curves.

    """
    u, inverse, counts = np.unique(Ee_vec, axis=0, return_inverse=True,
                                   return_counts=True)
    time_data = dict()
    time_data['Isys'] = sys_data['Isys'][inverse]
    time_data['Vsys'] = sys_data['Vsys'][inverse]
    time_data['Psys'] = sys_data['Psys'][inverse]
    time_data['Imp'] = sys_data['Imp'][inverse]
    time_data['Vmp'] = sys_data['Vmp'][inverse]
    time_data['Pmp'] = sys_data['Pmp'][inverse]
    time_data['Isc'] = sys_data['Isc'][inverse]
    time_data['Voc'] = sys_data['Voc'][inverse]
    time_data['FF'] = sys_data['FF'][inverse]

    return time_data


def calcACSystem(Ee_vec, Ee_mod, mod_data, NPT_dict, run_bpact=True, run_annual=False, save_bpact_freq=False):
    """
    Generate the system-level IV curves for the unique systems in a simulation. This function is for AC or MLPE systems.

    Parameters
    ----------
    Ee_vec : numpy.ndarray
        System level irradiance array.
    Ee_mod : numpy.ndarray
        3-D array containing the Irradiance at the cell level for all modules.
    mod_data : dict
        Dictionary containing module IV curves.
    NPT_dict : dict
        NPTs dictionary from the cell data dictionary generated by pvcell.
    run_bpact : bool, optional
        Flag to run bypass diode activation logic. The default is True.
    run_annual : bool, optional
        Flag to delete large bypass diode activation array in case of an annual simulation. The default is False.
    save_bpact_freq : bool, optional
        Flag to turn on saving bypass diode activation frequency input. The default is False.

    Returns
    -------
    sys_data : dict
        Dictionary containing system IV curves.

    """
    Ee_sys, _ = get_unique_Ee(Ee_vec, search_type='system')
    Ee_shp = Ee_sys.shape
    if run_bpact:
        BpDmp = np.zeros((Ee_shp[0], Ee_shp[1], Ee_shp[2], mod_data['BPDiode_Active_MPP'].shape[1]))
    Pmp = np.zeros(Ee_shp[0])
    if run_bpact:
        num_bpd_active = np.zeros(Ee_shp[0])
    else:
        num_bpd_active = 0
    for idx_0 in range(Ee_shp[0]):
        for idx_1 in range(Ee_shp[1]):
            # 1 String
            Ee_str1 = Ee_sys[idx_0, idx_1, :, :, :]
            # Extract mod IV curves
            str_in_mod = isin_nd(Ee_mod, Ee_str1)
            u, inverse, counts = np.unique(Ee_str1, axis=0, return_inverse=True,
                                           return_counts=True)
            Pmp_red = np.zeros((len(str_in_mod),))
            if run_bpact:
                num_bpd_active_red = np.zeros((len(str_in_mod),))
                BpDmp_red = np.zeros((len(str_in_mod), BpDmp.shape[-1]))
            for idx_red in range(str_in_mod.shape[0]):
                Ee_mod_red = Ee_mod[str_in_mod[idx_red], :, :]
                Ee_str_red = np.stack([Ee_mod_red], axis=0)
                Ee_str_red = np.stack([Ee_str_red], axis=0)
                Ee_sys_red = np.stack([Ee_str_red], axis=0)
                # String calculation
                str_data = calcStrings(Ee_str_red, Ee_mod, mod_data, NPT_dict, run_bpact=run_bpact)
                # System calculation
                sys_data1 = calcSystem(
                    Ee_sys_red, Ee_str_red, str_data, NPT_dict, run_bpact=run_bpact)
                Pmp_red[idx_red] = sys_data1['Pmp'][0]
                if run_bpact:
                    num_bpd_active_red[idx_red] = sys_data1['num_active_bpd'][0]
                    if save_bpact_freq:
                        BpDmp_red[idx_red, :] = sys_data1['Bypass_Active_MPP'][0, 0, :, :]
            Pmp_str = Pmp_red[inverse]
            Pmp[idx_0] += Pmp_str.sum()
            if run_bpact:
                num_bpd_active_str = num_bpd_active_red[inverse]
                num_bpd_active[idx_0] += num_bpd_active_str.sum()
                if save_bpact_freq:
                    BpDmp[idx_0, idx_1, :, :] = BpDmp_red[inverse, :]
    # Store results in a dict
    u, inverse, counts = np.unique(Ee_vec, axis=0, return_inverse=True,
                                   return_counts=True)
    Pmp_ts = Pmp[inverse]
    if run_bpact:
        num_bpd_active_ts = num_bpd_active[inverse]
        if save_bpact_freq:
            BpDmp_full = BpDmp[inverse, :, :, :]
        else:
            BpDmp_full = np.zeros(Ee_vec.shape)
    else:
        num_bpd_active_ts = 0
        BpDmp_full = np.zeros(Ee_vec.shape)
    sys_data = dict()
    sys_data['Isys'] = np.zeros(Pmp_ts.shape)
    sys_data['Vsys'] = np.zeros(Pmp_ts.shape)
    sys_data['Psys'] = np.zeros(Pmp_ts.shape)
    sys_data['Bypass_activation'] = np.zeros(Pmp_ts.shape)
    sys_data['Imp'] = np.zeros(Pmp_ts.shape)
    sys_data['Vmp'] = np.zeros(Pmp_ts.shape)
    sys_data['Pmp'] = Pmp_ts
    sys_data['Isc'] = np.zeros(Pmp_ts.shape)
    sys_data['Voc'] = np.zeros(Pmp_ts.shape)
    sys_data['FF'] = np.zeros(Pmp_ts.shape)
    sys_data['Bypass_Active_MPP'] = BpDmp_full
    sys_data['num_active_bpd'] = num_bpd_active_ts

    return sys_data


def calcSystem(Ee_sys, Ee_str, str_data, NPT_dict, run_bpact=True, run_annual=False):
    """
    Generate the system-level IV curves for the unique systems in a simulation. This function is for DC string systems.

    Parameters
    ----------
    Ee_sys : numpy.ndarray
        System level irradiance array.
    Ee_str : numpy.ndarray
        4-D array containing the Irradiance at the cell level for all modules in each string.
    str_data : dict
        Dictionary containing string IV curves.
    NPT_dict : dict
        NPTs dictionary from the cell data dictionary generated by pvcell.
    run_bpact : bool, optional
        Flag to run bypass diode activation logic. The default is True.
    run_annual : bool, optional
        Flag to delete large bypass diode activation array in case of an annual simulation. The default is False.

    Returns
    -------
    sys_data : dict
        Dictionary containing system IV curves.

    """
    I_sys_curves = []
    V_sys_curves = []
    P_sys_curves = []
    Bypass_str_curves = []
    for idx_sys in range(Ee_sys.shape[0]):

        # 1 String
        Ee_sys1 = Ee_sys[idx_sys]
        # Extract mod IV curves
        sys_in_str = isin_nd(Ee_str, Ee_sys1)
        Istr_red = str_data['Istring'][sys_in_str, :]
        Vstr_red = str_data['Vstring'][sys_in_str, :]
        if run_bpact:
            Bypass_substr_red = str_data['Bypassed_substr'][sys_in_str, :, :]
        else:
            Bypass_substr_red = np.nan
        u, inverse, counts = np.unique(Ee_sys1, axis=0, return_inverse=True,
                                       return_counts=True)
        # Expand for Str curves
        Istring = Istr_red[inverse, :]
        Vstring = Vstr_red[inverse, :]
        if run_bpact:
            Bypass_substr = Bypass_substr_red[inverse, :, :]
        else:
            Bypass_substr = np.nan
        pts = NPT_dict['pts'][0, :].reshape(NPT_dict['pts'].shape[1], 1)
        negpts = NPT_dict['negpts'][0, :].reshape(
            NPT_dict['negpts'].shape[1], 1)
        Npts = NPT_dict['Npts']
        # Run System Circuit model
        Isys, Vsys, bypassed_str = calcParallel_with_bypass(
            Istring, Vstring, Vstring.max(), Vstring.min(), negpts, pts, Npts, Bypass_substr, run_bpact=run_bpact)
        Psys = Isys * Vsys

        I_sys_curves.append(np.reshape(Isys, (1, len(Isys))))
        V_sys_curves.append(np.reshape(Vsys, (1, len(Vsys))))
        P_sys_curves.append(np.reshape(Psys, (1, len(Psys))))
        if run_bpact:
            Bypass_str_curves.append(np.reshape(bypassed_str, (1,
                                                               bypassed_str.shape[0],
                                                               bypassed_str.shape[1],
                                                               bypassed_str.shape[2],
                                                               bypassed_str.shape[3])))
        else:
            Bypass_str_curves.append(bypassed_str)

    I_sys_curves = np.concatenate(I_sys_curves, axis=0)
    V_sys_curves = np.concatenate(V_sys_curves, axis=0)
    P_sys_curves = np.concatenate(P_sys_curves, axis=0)
    if run_bpact:
        Bypass_str_curves = np.concatenate(Bypass_str_curves, axis=0)
    else:
        Bypass_str_curves = np.array(Bypass_str_curves)
    # If it is an annual simulation, delete the bypassed related variables
    if run_annual:
        del str_data['Bypassed_substr']

    Imp, Vmp, Pmp, Isc, Voc, FF, BpDmp, num_bpd_active = calcMPP_IscVocFFBPD(
        I_sys_curves, V_sys_curves, P_sys_curves, Bypass_str_curves, run_bpact=run_bpact, run_annual=run_annual)

    # Store results in a dict
    sys_data = dict()
    sys_data['Isys'] = I_sys_curves
    sys_data['Vsys'] = V_sys_curves
    sys_data['Psys'] = P_sys_curves
    if run_annual:
        sys_data['Bypass_activation'] = np.nan
        del Bypass_str_curves
    else:
        sys_data['Bypass_activation'] = Bypass_str_curves
    sys_data['Imp'] = Imp
    sys_data['Vmp'] = Vmp
    sys_data['Pmp'] = Pmp
    sys_data['Isc'] = Isc
    sys_data['Voc'] = Voc
    sys_data['FF'] = FF
    sys_data['Bypass_Active_MPP'] = BpDmp
    sys_data['num_active_bpd'] = num_bpd_active

    return sys_data



# -*- coding: utf-8 -*-
"""
This module contains configuration constants for PVMismatch, such as number of
points in IV curve to calculate, flag to use parallel processing and parallel
processing parameters. This module also contains some utility functions like
:func:`~pvmismatch.pvmismatch_lib.pvconstants.npinterpx()` and
:func:`~pvmismatch.pvmismatch_lib.pvconstants.get_series_cells()` are defined
here too.
"""

# TODO: move this to pvmismatch_lib/__init__.py

import numpy as np
import scipy.constants

# Constants
NPTS = 101  # number of I-V points to calculate
MODSIZES = [24, 72, 96, 128]  # list of possible number of cells per module
NUMBERCELLS = MODSIZES[2]  # default is 96-cell module
NUMBERMODS = 10  # default number of modules
NUMBERSTRS = 10  # default number of strings
EPS = np.finfo(np.float64).eps


def npinterpx(x, xp, fp):
    """
    Numpy interpolation function with linear extrapolation.

    Parameters
    ----------
    x : array_like
        The x-coordinates of the interpolated values.

    xp : 1-D sequence of floats
        The x-coordinates of the data points, must be increasing.

    fp : 1-D sequence of floats
        The y-coordinates of the data points, same length as `xp`.

    Returns
    -------
    y : {float, ndarray}
        The interpolated values, same shape as `x`.

    Raises
    ------
    ValueError
        If `xp` and `fp` have different length
    """
    y = np.interp(x, xp, fp)
    # extrapolate left
    left = x < xp[0]
    xleft = x[left]
    yleft = fp[0] + (xleft - xp[0]) / (xp[1] - xp[0]) * (fp[1] - fp[0])
    y[left] = yleft
    # extrapolate right
    right = x > xp[-1]
    xright = x[right]
    yright = fp[-1] + (xright - xp[-1]) / (xp[-2] - xp[-1]) * (fp[-2] - fp[-1])
    y[right] = yright
    return y


class PVconstants(object):
    """
    Class for configuration constants

    :param npts: number of points in IV curve
    :type npts: int
    """
    # hard constants
    k = scipy.constants.k  #: [J/K] Boltzmann constant
    q = scipy.constants.e  #: [Coloumbs] elementary charge
    E0 = 1000.  #: [W/m^2] irradiance of 1 sun
    T0 = 298.15  #: [K] reference temperature

    def __init__(self, npts=NPTS):
        self._npts = None
        self.pts = None
        """array of points with decreasing spacing from exactly zero to one"""
        self.negpts = None
        """array of points with increasing spacing from one to ``0.1/npts``"""
        self.Imod_pts = None
        """array of points with increasing spacing from exactly zero to one"""
        self.Imod_pts_sq = None
        """array of points from EPS to 1 with very close spacing near EPS"""
        self.Imod_negpts = None
        """array of points with decreasing spacing from one to  ``0.1/npts``"""
        self.Vmod_q4pts = None
        """array of points with decreasing spacing from ``0.1/npts`` to one"""
        # call property setter
        self.npts = npts  #: number of points in IV curves

    @property
    def npts(self):
        """number of points in IV curves"""
        return self._npts

    @npts.setter
    def npts(self, npts):
        # set number of points in IV curve(s)
        self._npts = npts  # number of points in IV curves
        # point spacing from 0 to 1, used for Vcell, Vmod, Vsys and Istring
        # decrease point spacing as voltage approaches Voc by using logspace
        pts = (11. - np.logspace(np.log10(11.), 0., self._npts)) / 10.
        pts[0] = 0.  # first point must be exactly zero
        self.pts = pts.reshape((self._npts, 1))
        negpts = (11. - np.logspace(np.log10(11. - 1. / float(self._npts)),
                                    0., self._npts)) / 10.
        negpts = negpts.reshape((self._npts, 1))
        self.Imod_negpts = 1 + 1. / float(self._npts) / 10. - negpts
        self.negpts = np.flipud(negpts)  # reverse the order
        # shift and concatenate pvconst.negpts and pvconst.pts
        # so that tight spacing is around MPP and RBD
        self.Imod_pts = 1 - np.flipud(self.pts)
        self.Imod_pts_sq = self.Imod_pts**2 + EPS
        self.Vmod_q4pts = np.flipud(self.Imod_negpts)

    def __str__(self):
        return '<PVconstants(npts=%d)>' % self.npts

    def __repr__(self):
        return str(self)

    def calcSeries(self, I, V, meanIsc, Imax):
        """
        Calculate IV curve for cells and substrings in series given current and
        voltage in increasing order by voltage, the average short circuit
        current and the max current at the breakdown voltage.

        :param I: cell or substring currents [A]
        :param V: cell or substring voltages [V]
        :param meanIsc: average short circuit current [A]
        :param Imax: maximum current [A]
        :return: current [A] and voltage [V] of series
        """
        # make sure all inputs are numpy arrays, but don't make extra copies
        I = np.asarray(I)  # currents [A]
        V = np.asarray(V)  # voltages [V]
        meanIsc = np.asarray(meanIsc)  # mean Isc [A]
        Imax = np.asarray(Imax)  # max current [A]
        # create array of currents optimally spaced from mean Isc to  max VRBD
        Ireverse = (Imax - meanIsc) * self.Imod_pts_sq + meanIsc
        # range of currents in forward bias from 0 to mean Isc
        Iforward = meanIsc * self.pts
        Imin = np.minimum(I.min(), 0.)  # minimum cell current, at most zero
        # range of negative currents in the 4th quadrant from min current to 0
        Iquad4 = Imin * self.Imod_negpts
        # create range for interpolation from forward to reverse bias
        Itot = np.concatenate((Iquad4, Iforward, Ireverse), axis=0).flatten()
        Vtot = np.zeros((3 * self.npts,))
        # add up all series cell voltages
        for i, v in zip(I, V):
            # interp requires x, y to be sorted by x in increasing order
            Vtot += npinterpx(Itot, np.flipud(i), np.flipud(v))
        return np.flipud(Itot), np.flipud(Vtot)

    def calcParallel(self, I, V, Vmax, Vmin, Voc=None):
        """
        Calculate IV curve for cells and substrings in parallel.

        :param I: currents [A]
        :type: I: list, :class:`numpy.ndarray`
        :param V: voltages [V]
        :type: V: list, :class:`numpy.ndarray`
        :param Vmax: max voltage limit, should be max Voc [V]
        :param Vmin: min voltage limit, could be zero or Vrbd [V]
        :param Voc: (``None``) open circuit voltage [V]
        """
        if Voc is None:
            Voc = Vmax
        I, V = np.asarray(I), np.asarray(V)
        Vmax = np.asarray(Vmax)
        Vmin = np.asarray(Vmin)
        Voc = np.asarray(Voc)
        Vff = Voc
        delta_Voc = Vmax - Voc
        if np.isclose(delta_Voc, 0):
            Vff = 0.8 * Voc
            delta_Voc = 0.2 * Voc
        elif delta_Voc < 0:
            Vff = Vmax
            delta_Voc = -delta_Voc
        Vquad4 = Vff + delta_Voc * self.Vmod_q4pts
        Vreverse = Vmin * self.negpts
        Vforward = Vff * self.pts
        Vtot = np.concatenate((Vreverse, Vforward, Vquad4), axis=0).flatten()
        Itot = np.zeros((3 * self.npts,))
        for i, v in zip(I, V):
            Itot += npinterpx(Vtot, v, i)
        return Itot, Vtot


def Vdiode(Icell, Vcell, Rs):
    """
    Calculate Vdiode from current, voltage and series resistance.

    :param Icell: cell current [A]
    :param Vcell: cell voltage [V]
    :param Rs: cell series resistance [:math:`\Omega`]
    :return: diode voltage [V]
    """
    return Vcell + Icell * Rs


def Idiode(Isat, Vdiode, Vt, n):
    """
    Calculate diode current using `Shockley diode model`_.

    :param Isat: diode saturation current [A]
    :param Vdiode: diode voltage [V]
    :param Vt: thermal voltage [V]
    :param n: diode ideality factor
    :return: diode current

    .. LaTeX math directives need double backslashes in comments

    .. note::
       Thermal voltage is defined as ...

       .. math::
           V_t = \\frac {k_B * T} {q_e}

    .. _Shockley diode model: https://en.wikipedia.org/wiki/Diode_modelling#Shockley_diode_model
    """
    return Isat * (np.exp(Vdiode / n / Vt) - 1.)


def Ishunt(Vdiode, Rsh):
    return Vdiode / Rsh


def Igen(Aph, Ee, Isc0):
    return Aph * Ee * Isc0


def get_series_cells(cell_pos_column, prev_col=None):
    """
    Get the sequence of series cells between parallel crossties.

    :param cell_pos_column: column in cell position pattern
    :param prev_col: previous column in cell position pattern
    :return: indices of series cells
    """
    series_cells = []  # empty list of indices of cells in series
    # if the previous column is specified, find the indices of cells in the
    # current column that correspond to cells between parallel crossties in the
    # previous column
    if prev_col:
        cell_pos_column = zip(prev_col, cell_pos_column)
    for cell in cell_pos_column:
        if prev_col:
            cell, next_col = cell
        else:
            next_col = None
        # noinspection PyTypeChecker
        if cell['crosstie'] == True:
            yield series_cells
            series_cells = []
        # if the next column is specified, return the cell indices that
        # correspond to the previous column since they must be the same
        if next_col:
            cell_idx = next_col['idx']
        else:
            # noinspection PyTypeChecker
            cell_idx = cell['idx']
        series_cells.append(cell_idx)
    yield series_cells
